﻿// Copyright (C) 2011 Oliver Sturm <oliver@oliversturm.com>. All rights reserved.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;
using System.Reflection;

namespace Tracing.Complete {
  class Program {
    static void Main(string[] args) {
      Expression<Func<int, int>> squareExp = x => x * x;
      var square = squareExp.Compile( );
      square = BoringTrace(square);
      Console.WriteLine(square(7));

      Expression<Func<int, int, int, int>> add3 = (x, y, z) => x + y + z;
      var add3_ = Trace(add3);
      Console.WriteLine(add3_(10, 20, 30));

      Console.WriteLine(SumOfOddNumbers(20));
      Console.WriteLine(TracedSumOfOddNumbers(20));
    }

    static Func<T, R> BoringTrace<T, R>(Func<T, R> source) {
      return x => {
        Console.WriteLine("--> ({0}) with {1}", source, x);
        var result = source(x);
        Console.WriteLine("<-- Returning ({1}) from ({0})", source, result);
        return result;
      };
    }

    static int SumOfOddNumbers(int cutoff) {
      return Enumerable.Range(1, cutoff).
        Where(x => x % 2 == 1).
        Aggregate(0, (r, v) => r + v);
    }

    static int TracedSumOfOddNumbers(int cutoff) {
      // Prize question here: why do I have to give the types 
      // explicitly to make the compiler happy on the Trace call?
      // After all, when I pass the same lambda to Where, I don't
      // have type information and obviously the compiler is quite
      // happy to infer it there...
      // (There is an answer, think about it!)
      return Enumerable.Range(1, cutoff).
        Where(Trace<int, bool>(x => x % 2 == 1)).
        Aggregate(0, Trace<int, int, int>((r, v) => r + v));
    }

    static IEnumerable<Expression> ArrayInitHelper(string firstVal, IEnumerable<Expression> exps) {
      yield return Expression.Constant(firstVal);
      foreach (Expression exp in exps)
        yield return Expression.Convert(exp, typeof(object));
    }

    static string GetLogFormat(Expression[] parameters) {
      var logFormatBuilder = new StringBuilder("--> Calling ({0}) with (");
      int num = 1;
      foreach (Expression parameter in parameters) {
        if (num > 1)
          logFormatBuilder.Append(", ");
        logFormatBuilder.Append("{" + (num++) + "}");
      }
      logFormatBuilder.Append(")");
      return logFormatBuilder.ToString( );
    }

    static Expression GetTraceExpression<R>(Expression exp, Expression[] parameters) {
      var result = Expression.Variable(typeof(R), "r");
      var writeLineFunction = typeof(Console).GetMethod("WriteLine",
        BindingFlags.FlattenHierarchy | BindingFlags.Public | BindingFlags.Static, null,
        new Type[] { typeof(string), typeof(object[]) }, null);
      var thisExpStr = exp.ToString( );
      string logFormat = GetLogFormat(parameters);

      return Expression.Block(
          new[] { result },
          Expression.Call(writeLineFunction,
            Expression.Constant(logFormat),
            Expression.NewArrayInit(typeof(object),
              ArrayInitHelper(thisExpStr, parameters))),
          Expression.Assign(result, Expression.Invoke(exp, parameters)),
          Expression.Call(writeLineFunction,
            Expression.Constant("<-- Returning ({1}) from ({0})"),
            Expression.NewArrayInit(typeof(object),
              Expression.Constant(thisExpStr),
              Expression.Convert(result, typeof(object)))),
          result
        );
    }

    static Func<P1, R> Trace<P1, R>(Expression<Func<P1, R>> exp) {
      var parameters = new ParameterExpression[] {
        Expression.Parameter(typeof(P1), "p1")};

      var traceExpression = GetTraceExpression<R>(exp, parameters);

      return
        Expression.Lambda<Func<P1, R>>(
          traceExpression,
          parameters).Compile( );
    }

    static Func<P1, P2, R> Trace<P1, P2, R>(Expression<Func<P1, P2, R>> exp) {
      var parameters = new ParameterExpression[] {
        Expression.Parameter(typeof(P1), "p1"),
        Expression.Parameter(typeof(P2), "p2")};

      var traceExpression = GetTraceExpression<R>(exp, parameters);

      return
        Expression.Lambda<Func<P1, P2, R>>(
          traceExpression,
          parameters).Compile( );
    }

    static Func<P1, P2, P3, R> Trace<P1, P2, P3, R>(Expression<Func<P1, P2, P3, R>> exp) {
      var parameters = new ParameterExpression[] {
        Expression.Parameter(typeof(P1), "p1"),
        Expression.Parameter(typeof(P2), "p2"),
        Expression.Parameter(typeof(P3), "p3")};

      var traceExpression = GetTraceExpression<R>(exp, parameters);

      return
        Expression.Lambda<Func<P1, P2, P3, R>>(
          traceExpression,
          parameters).Compile( );
    }

  }
}
